
/*
 * The intention is to open the device with the parameters configured in the
 * few accepted switches, and read/write the input filename to the audio
 * device. This will allow some testing of the import/export facilities.
 */

#include <sys/types.h>
#include <fcntl.h>
#include <sys/time.h>
#include <unistd.h>
#ifdef __FreeBSD__
#include <stdlib.h>
#else
#include <malloc.h>
#endif

#include "slabDAccess.h"

#ifndef NULL
#define NULL 0
#endif

#define usage "<-m[ono]> <-8> <-D(ual)> <-r rate> <-d audioDev> <-f AU_file>"

#ifdef sun
#define DefaultDev "/dev/null"
#include <sys/ioccom.h>
#include <sun/audioio.h>
#include <fcntl.h>

audio_info_t audioInfo;
#else /* sun vs LINUX */
#define DefaultDev "/dev/audio"
#ifdef LINUX
#include <sys/soundcard.h>
audio_buf_info bufferInfo;
#endif
#endif

char *audioDev;

int channels = 1, resolution = 16, sampleRate = 44100;
int inputFD = 0, outputFD = -1, duplexInputFD = -1;
int sbDuplex = 0;
int readCount;
int dualOpen = 0;
int loopCount = 1;
int fileFormat = 0, fragCount = 0;

int dataBufferSize = 8192;
char *dataBuffer; /* [8192]; */
char nullBuffer[8192];
char *inputFileName;

nbeHeader slabFileHdr;

fd_set			Input_fds, Output_fds, Except_fds;
int setCount;

main(argc, argv)
int argc;
char *argv[];
{
	audioDev = DefaultDev;
	/*
	 * Switches are channels, speed, resolution
	 */
	checkSwitches(argc, argv);

#ifdef DEBUG
	printf("chans = %i, res = %i, rate = %i, ",
		channels, resolution, sampleRate);
	printf("inputFD = %i, outputFD = %i, %s\n",
		inputFD, outputFD, audioDev);
#endif

	/*
	 * We now have all our parameters, need to configure the audioDev for
	 * our parameters, and read inputFD to outputFD.
	 */
	initAudioDevice();

	FD_ZERO(&Input_fds);
	FD_ZERO(&Output_fds);

	while (loopCount-- > 0)
	{
		while ((readCount = readData(inputFD, dataBuffer, 8192)) > 0)
		{
			FD_SET(outputFD, &Output_fds);

			if (duplexInputFD != -1)
			{
				FD_SET(duplexInputFD, &Input_fds);

				setCount = select(FD_SETSIZE, &Input_fds, &Output_fds,
					&Except_fds, (struct timeval *) NULL);

				if (FD_ISSET(duplexInputFD, &Input_fds))
				{
					printf("Read data available\n");
					/*
					 * If we are sbDuplex the input channel is only half the
					 * resolution. Consequently, to stay in sync we only read
					 * half the amount of data.....
					 */
					if (sbDuplex)
						read(duplexInputFD, nullBuffer, 4096);
					else
						read(duplexInputFD, nullBuffer, 8192);
					printf("Read data done\n");
					setCount--;
				}
				if (FD_ISSET(outputFD, &Output_fds))
				{
					printf("Write space available\n");

					if (write(outputFD, dataBuffer, readCount) < 0)
					{
						printf("Write Failed\n");
						exit(-6);
					}
					setCount--;
				}
				if (setCount != 0)
				{
					printf("Huh?\n");
					FD_ZERO(&Input_fds);
					FD_ZERO(&Output_fds);
				}
			} else {
				if (write(outputFD, dataBuffer, readCount) < 0)
				{
					printf("Write Failed\n");
					exit(-6);
				}
			}
		}
	}

	close(inputFD);
	close(outputFD);
	if (duplexInputFD != -1)
		close(duplexInputFD);
	exit(0);
}

checkSwitches(argc, argv)
int argc;
char *argv[];
{
	int i = 1;

	if (argc == 1)
	{
		printf("%s: %s\n", argv[0], usage);
		exit(-1);
	}

	while ((i < argc) && (argv[i][0] == '-'))
	{
		if (argv[i][1] == 'C')
			sbDuplex = 1;
		if (argv[i][1] == 's')
			fileFormat = 1;
		if (argv[i][1] == 'M')
			dualOpen = -1;
		if (argv[i][1] == 'D')
			dualOpen = 1;
		if (argv[i][1] == 'l')
		{
			loopCount = 1024;
			/*
			 * see if we have been given a loopcount
			 */
			if (argv[i + 1][0] != '-')
			{
				loopCount = atoi(argv[++i]);
			}
		}
		if (argv[i][1] == 'm')
			channels = 0;
		if (argv[i][1] == '8')
			resolution = 8;
		if ((argv[i][1] == 'r') && (i < argc))
		{
			sampleRate = atoi(argv[i + 1]);
			i++;
		}
		/*
		 * We have 2 alternatives here. Firstly try a DUP() on the outputFD
		 * after it is opened RW. Alternatively open once RO, and again WO.
		 */
		if ((argv[i][1] == 'd') && (i < argc))
		{
			audioDev = argv[i + 1];
			i++;
		}
		if ((argv[i][1] == 'f') && (i < argc))
		{
			inputFileName = argv[i + 1];
			if ((inputFD = open(argv[i + 1], O_RDONLY)) < 0)
			{
				printf("%s: Failed to input file %s\n", argv[0], argv[i + 1]);
				exit(-3);
			}
#ifdef DEBUG
			printf("Opened input file %s\n", argv[i + 1]);
#endif
			i++;
		}

		i++;
	}

	/*
	 * If we only have one parameter, and it is not a switch, assume it is the
	 * file to be played at the set values: this allows "slabPlayer <file>".
	 */
	if (i < argc)
	{
		inputFileName = argv[i];
		if ((inputFD = open(argv[i], O_RDONLY)) < 0)
		{
			printf("%s: Failed open to input device %s\n", argv[0], argv[i]);
				exit(-4);
		}
#ifdef DEBUG
		printf("Opened input file %s\n", argv[i]);
#endif
		i++;
	}

	if (dualOpen == 0)
	{
		if ((outputFD = open(audioDev, O_RDWR)) < 0)
		{
			printf("%s: Failed to open device %s\n",
				argv[0], audioDev);
			exit(-2);
		}
		duplexInputFD = fcntl(outputFD, F_DUPFD, outputFD);
	} else if (dualOpen == 1) {
		if ((outputFD = open(audioDev, O_WRONLY)) < 0)
		{
			printf("%s: Failed to open device %s for output\n",
				argv[0], audioDev);
			exit(-2);
		}
		if ((duplexInputFD = open(audioDev, O_RDONLY)) < 0)
		{
			printf("%s: Failed to open device %s for input\n",
				argv[0], audioDev);
			exit(-2);
		}
	} else {
		/*
		 * This is going to be outbound data only
		 */
		if ((outputFD = open(audioDev, O_WRONLY)) < 0)
		{
			printf("%s: Failed to open device %s\n",
				argv[0], audioDev);
			exit(-2);
		}
		duplexInputFD == -1;
	}
	if (fileFormat)
	{
		initSLabFile(inputFD);
	}
	dataBuffer = (char *) malloc(dataBufferSize);
}

/*
 * Called once when the device is opened by the engine. Initialises our fd to
 * 44100 16 bits stereo operation.
 */
int
initAudioDevice()
{
#ifdef LINUX
	int trackfile, i, j, results, data, mode, index;

	/*
	 * The device is basically just opened for the first call of this routine.
	 * This is then called several times afterwards - each time we start the
	 * engine (that is a tape start, not an engine restart). This may be
	 * problematic.
	 *
	 * Set the fragments.
	 * In this order set bits, channels, speed.
	 */
	if ((duplexInputFD != -1) && (ioctl(outputFD, SNDCTL_DSP_SETDUPLEX, 0) < 0))
		printf("Failed to set duplex on device\n");

	/*
	 * 2 Fragments. The fragment size is 8k.
	 */
	data = 0x0002000d;

	/*
	 * Set up the partial write operations.
	 *
	 * THIS SHOULD NOW BE DEFAULT BEHAVIOUR
	 *
	 */
	if ((results = ioctl(outputFD,
		SNDCTL_DSP_SETFRAGMENT, &data))
		== 0)
#ifdef DEBUG
		printf("set audio fragment: %x\n", data);
	else
		printf("set audio fragment failed: %x\n", data);
#else
		;
#endif

	/*
	 * Check capabilities.
	 */
	if ((results =
		ioctl(outputFD, SNDCTL_DSP_GETCAPS, &data)) == 0)
	{
		/*
		 * Take this out, we only need to check the options here.
		 *
		 */
		if (data & DSP_CAP_DUPLEX)
		{
			if (ioctl(outputFD,
				DSP_CAP_DUPLEX, &data) < 0)
#ifdef DEBUG
				printf("Set to Duplex\n");
			else
				printf("Failed to set Duplex\n");
#else
				;
#endif
		}
	}

	/*
	 * Get current bits (0x10 = 16 bit, 0x8 = 8 bit
	 */
	data = 0;
	if ((results =
		ioctl(outputFD, SNDCTL_DSP_GETFMTS, &data)) == 0)
#ifdef DEBUG
		printf("Current audio formats: %i\n", data);
#else
		;
#endif
	else
		printf("Get sound format failed: %i\n", results);

	/*
	 * Set to 16 bit only
	 */
	data = resolution;
	if ((results =
		ioctl(outputFD, SOUND_PCM_WRITE_BITS, &data))
		== 0)
#ifdef DEBUG
		printf("Set write resolution: %i\n", data);
#else
		;
#endif
	else
		printf("Set resolution failed: %i\n", results);

	if (sbDuplex)
		data = 8;
	else
		data = resolution;

	if ((results =
		ioctl(outputFD, SOUND_PCM_READ_BITS, &data))
		== 0)
#ifdef DEBUG
		printf("Set read resolution: %i\n", data);
#else
		;
#endif
	else
		printf("Set resolution failed: %i\n", results);

	/*
	 * Set to stereo
	 */
	data = channels;
	mode = SNDCTL_DSP_STEREO;
	if ((results =
		ioctl(outputFD, mode, &data)) == 0)
#ifdef DEBUG
		printf("Set to stereo: %i\n", data);
#else
		;
#endif
	else
		printf("Set stereo failed: %i\n", results);

	/*
	 * Set to AUDIO_RATE Hz
	 */
	data = sampleRate;
	if ((results =
		ioctl(outputFD, SOUND_PCM_WRITE_RATE, &data))
		== 0)
#ifdef DEBUG
		printf("Set audio write rate to %i\n", data);
#else
		;
#endif
	else
		printf("Set line frequency failed: %i\n", results);

	/*
	 * Write rate may be different to read rate, and physical IO speeds
	 * may be different from disk recorded rate.....
	 */
	data = sampleRate;
	if ((results =
		ioctl(outputFD, SOUND_PCM_READ_RATE, &data))
		== 0)
#ifdef DEBUG
		printf("Set audio read rate to %i\n", data);
#else
		;
#endif
	else
		printf("Set line frequency failed: %i\n", results);

#endif /* LINUX */

#ifdef sun
	int results = 0;

	AUDIO_INITINFO(&audioInfo);

	if ((results =
		ioctl(outputFD, AUDIO_GETDEV, &results)) < 0)
	{
		perror("Failed to set audio dev info");
		return(-2);
	}

#ifdef DEBUG
	switch(results) {
		default:
			printf("Device unknown\n");
			break;
		case 1:
			printf("Device AMD\n");
			break;
		case 2:
			printf("Device DBRI\n");
			break;
		case 3:
			printf("Device CODEC\n");
			break;
	}
#endif

	/*
	 * Use zero gain, for testing
	 */
	audioInfo.play.gain = 0;
	audioInfo.record.gain = 0;

	/*
	 * These are the preferred parameters
	 */
	audioInfo.play.sample_rate = sampleRate;
	audioInfo.record.sample_rate = sampleRate;

	audioInfo.play.encoding = AUDIO_ENCODING_LINEAR;
	audioInfo.record.encoding = AUDIO_ENCODING_LINEAR;

	audioInfo.play.channels = channels + 1;
	audioInfo.record.channels = channels + 1;

	audioInfo.play.precision = resolution;
	audioInfo.record.precision = resolution;

	audioInfo.play.port = AUDIO_LINE_IN;
	audioInfo.record.port = AUDIO_LINE_IN;

	if ((results =
		ioctl(outputFD, AUDIO_SETINFO, &audioInfo))
		< 0)
	{
		/*
		 * We may only have AMD capabilities. Check it out.
		 */
		audioInfo.play.sample_rate = 8000;
		audioInfo.record.sample_rate = 8000;

		audioInfo.play.encoding = AUDIO_ENCODING_ALAW;
		audioInfo.record.encoding = AUDIO_ENCODING_ALAW;

		audioInfo.play.channels = 1;
		audioInfo.record.channels = 1;

		audioInfo.play.precision = 8;
		audioInfo.record.precision = 8;

		audioInfo.play.port = 1;
		audioInfo.record.port = 1;

		if ((results = ioctl(outputFD, AUDIO_SETINFO,
			&audioInfo)) < 0)
		{
			/*
			 * This this happened then we do not have any capabilities
			 */
			printf("Failed to get audio info\n");
			return(-2);
		}
	}

#ifdef DEBUG
	printf("play sampleRate %i\n", audioInfo.play.sample_rate);
	printf("record sampleRate %i\n", audioInfo.record.sample_rate);

	printf("play encoding %i\n", audioInfo.play.encoding);
	printf("record encoding %i\n", audioInfo.record.encoding);

	printf("play channels %i\n", audioInfo.play.channels);
	printf("record channels %i\n", audioInfo.record.channels);

	printf("play precision %i\n", audioInfo.play.precision);
	printf("record precision %i\n", audioInfo.record.precision);

	printf("play port %i\n", audioInfo.play.port);
	printf("record port %i\n", audioInfo.record.port);

	printf("play avail_ports %i\n", audioInfo.play.avail_ports);
	printf("record avail_ports %i\n", audioInfo.record.avail_ports);
#endif /* DEBUG */
	/*
	 * Move the required variables over to the controlBuffer, so that the
	 * code can converge to these values
	 */
#endif /* sun */

	return(0);
}

/*
 * This will load data into the buffer from the file, depending on the format.
 * Note that SLab edit file format will only work for cut/copy files, not for
 * fade/reverse/merge, since the fragment sequences may not be in order.
 */
readData(fd, buffer, count)
char *buffer;
{
	int readCount;

	switch (fileFormat)
	{
		case 1:
			/*
			 * SLab V1 edit file format
			 */
			if (--fragCount < 0)
			{
				initSLabFile(fd);
			}
			readCount = readFrag(fd, buffer, count);
			break;
		case 0:
			/*
			 * Raw file format
			 */
			if ((readCount = read(fd, buffer, count)) < count)
				lseek(inputFD, (long) 0, SEEK_SET);
			break;
	}
	return(readCount);
}

readFrag(fd, buffer, count)
{
	int fragSize, fragOffset;

	switch (fileFormat)
	{
		case 1:
			read(fd, &fragSize, sizeof(int));
			fragSize = fragSize * sizeof(short);
			read(fd, &fragOffset, sizeof(int));

			if (fragSize > dataBufferSize)
			{
				if (dataBuffer)
					free(dataBuffer);
				dataBufferSize = fragSize;
				dataBuffer = (char *) malloc(dataBufferSize);
			}

			read(fd, dataBuffer, fragSize);

			return(fragSize);
		default:
			printf("Unknown SLab edit file version %i\n", fileFormat);
			exit(-7);
			break;
	}
}

initSLabFile(fd)
{
	lseek(fd, (long) 0, SEEK_SET);

	read(fd, &fileFormat, sizeof(int));

	switch (fileFormat)
	{
		case 1:
			if (slabReadV1Header(fd, &slabFileHdr) < 0)
			{
				printf("Unknown SLab edit file version %i\n", fileFormat);
				exit(-9);
			}
			fragCount = slabFileHdr.fragments;
			break;
		default:
			printf("Unknown SLab edit file version %i\n", fileFormat);
			exit(-8);
			break;
	}
}

slabReadV1Header(nbeFD, header)
nbeHeader *header;
{

#ifdef DEBUG
	printf("slabReadV1Header(%i)\n", nbeFD);
#endif
	read(nbeFD, &header->track, sizeof(int));
	read(nbeFD, &header->samples, sizeof(int));
	read(nbeFD, &header->fragments, sizeof(int));
	read(nbeFD, &header->offset, sizeof(int));

	read(nbeFD, &header->fnl, sizeof(int));
	if ((header->fnl >= 128) || (header->fnl < 0))
		return(-1);
	read(nbeFD, &header->fn[0], header->fnl);
	header->fn[header->fnl] = '\0';

	read(nbeFD, &header->descl, sizeof(int));
	if ((header->descl >= 128) || (header->fnl < 0))
		return(-1);
	read(nbeFD, &header->desc[0], header->descl);
	header->desc[header->descl] = '\0';

	return(0);
}

